@@ -12,13 +12,23 @@ import errno
 import signal
 import sys
 import time
+try:
+    import subprocess
+except ImportError:
+    from paste.util import subprocess24 as subprocess
 from command import Command, BadCommand
 from paste.deploy import loadapp, loadserver
 import threading
 import atexit
+import logging
+import ConfigParser
 
 MAXFD = 1024
 
+class DaemonizeException(Exception):
+    pass
+
+
 class ServeCommand(Command):
 
     min_args = 0
@@ -52,10 +62,11 @@ class ServeCommand(Command):
                       dest='server_name',
                       metavar='SECTION_NAME',
                       help="Use the named server as defined in the configuration file (default: main)")
-    parser.add_option('--daemon',
-                      dest="daemon",
-                      action="store_true",
-                      help="Run in daemon (background) mode")
+    if hasattr(os, 'fork'):
+        parser.add_option('--daemon',
+                          dest="daemon",
+                          action="store_true",
+                          help="Run in daemon (background) mode")
     parser.add_option('--pid-file',
                       dest='pid_file',
                       metavar='FILENAME',
@@ -72,13 +83,18 @@ class ServeCommand(Command):
                       dest='reload_interval',
                       default=1,
                       help="Seconds between checking files (low number can cause significant CPU usage)")
+    parser.add_option('--monitor-restart',
+                      dest='monitor_restart',
+                      action='store_true',
+                      help="Auto-restart server if it dies")
     parser.add_option('--status',
                       action='store_true',
                       dest='show_status',
                       help="Show the status of the (presumably daemonized) server")
 
+
     if hasattr(os, 'setuid'):
-        # I don't think these are availble on Windows
+        # I don't think these are available on Windows
         parser.add_option('--user',
                           dest='set_user',
                           metavar="USERNAME",
@@ -94,11 +110,12 @@ class ServeCommand(Command):
                       help='Stop a daemonized server (given a PID file, or default paster.pid file)')
 
 
-    _scheme_re = re.compile(r'^[a-zA-Z0-9_-]+:')
+    _scheme_re = re.compile(r'^[a-z][a-z]+:', re.I)
 
     default_verbosity = 1
 
     _reloader_environ_key = 'PYTHON_RELOADER_SHOULD_RUN'
+    _monitor_environ_key = 'PASTE_MONITOR_SHOULD_RUN'
 
     possible_subcommands = ('start', 'stop', 'restart', 'status')
     def command(self):
@@ -175,18 +192,59 @@ class ServeCommand(Command):
             server_spec = app_spec
         base = os.getcwd()
 
-        if self.options.daemon:
-            self.daemonize()
+        if getattr(self.options, 'daemon', False):
+            if not self.options.pid_file:
+                self.options.pid_file = 'paster.pid'
+            if not self.options.log_file:
+                self.options.log_file = 'paster.log'
+
+        # Ensure the log file is writeable
+        if self.options.log_file:
+            try:
+                writeable_log_file = open(self.options.log_file, 'a')
+            except IOError, ioe:
+                msg = 'Error: Unable to write to log file: %s' % ioe
+                raise BadCommand(msg)
+            writeable_log_file.close()
+
+        # Ensure the pid file is writeable
+        if self.options.pid_file:
+            try:
+                writeable_pid_file = open(self.options.pid_file, 'a')
+            except IOError, ioe:
+                msg = 'Error: Unable to write to pid file: %s' % ioe
+                raise BadCommand(msg)
+            writeable_pid_file.close()
+
+        if getattr(self.options, 'daemon', False):
+            try:
+                self.daemonize()
+            except DaemonizeException, ex:
+                if self.verbose > 0:
+                    print str(ex)
+                return
+
+        if (self.options.monitor_restart
+            and not os.environ.get(self._monitor_environ_key)):
+            return self.restart_with_monitor()
 
         if self.options.pid_file:
             self.record_pid(self.options.pid_file)
 
         if self.options.log_file:
-            stdout_log = LazyWriter(self.options.log_file)
+            stdout_log = LazyWriter(self.options.log_file, 'a')
             sys.stdout = stdout_log
             sys.stderr = stdout_log
-            # @@: Should we also redirect logging-based logs to this file?
-            # e.g., the root logger?
+            logging.basicConfig(stream=stdout_log)
+
+        log_fn = app_spec
+        if log_fn.startswith('config:'):
+            log_fn = app_spec[len('config:'):]
+        elif log_fn.startswith('egg:'):
+            log_fn = None
+        if log_fn:
+            log_fn = os.path.join(base, log_fn)
+            self.logging_file_config(log_fn)
 
         server = self.loadserver(server_spec, name=server_name,
                                  relative_to=base, global_conf=vars)
@@ -217,6 +275,12 @@ class ServeCommand(Command):
                 **kw)
 
     def daemonize(self):
+        pid = live_pidfile(self.options.pid_file)
+        if pid:
+            raise DaemonizeException(
+                "Daemon is already running (PID: %s from PID file %s)"
+                % (pid, self.options.pid_file))
+
         if self.verbose > 0:
             print 'Entering daemon mode'
         pid = os.fork()
@@ -234,7 +298,7 @@ class ServeCommand(Command):
 
         # @@: Should we set the umask and cwd now?
 
-        import resource		# Resource usage information.
+        import resource  # Resource usage information.
         maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
         if (maxfd == resource.RLIM_INFINITY):
             maxfd = MAXFD
@@ -242,23 +306,18 @@ class ServeCommand(Command):
         for fd in range(0, maxfd):
             try:
                 os.close(fd)
-            except OSError:	# ERROR, fd wasn't open to begin with (ignored)
+            except OSError:  # ERROR, fd wasn't open to begin with (ignored)
                 pass
 
         if (hasattr(os, "devnull")):
             REDIRECT_TO = os.devnull
         else:
             REDIRECT_TO = "/dev/null"
-        os.open(REDIRECT_TO, os.O_RDWR)	# standard input (0)
+        os.open(REDIRECT_TO, os.O_RDWR)  # standard input (0)
         # Duplicate standard input to standard output and standard error.
-        os.dup2(0, 1)			# standard output (1)
-        os.dup2(0, 2)			# standard error (2)
+        os.dup2(0, 1)  # standard output (1)
+        os.dup2(0, 2)  # standard error (2)
         
-        if not self.options.pid_file:
-            self.options.pid_file = 'paster.pid'
-        if not self.options.log_file:
-            self.options.log_file = 'paster.log'
-
     def record_pid(self, pid_file):
         pid = os.getpid()
         if self.verbose > 1:
@@ -315,23 +374,47 @@ class ServeCommand(Command):
         return 0
 
     def restart_with_reloader(self):
+        self.restart_with_monitor(reloader=True)
+
+    def restart_with_monitor(self, reloader=False):
         if self.verbose > 0:
-            print 'Starting subprocess with file monitor'
+            if reloader:
+                print 'Starting subprocess with file monitor'
+            else:
+                print 'Starting subprocess with monitor parent'
         while 1:
-            args = [sys.executable] + sys.argv
-            if sys.platform == "win32":
-                args = ['"%s"' % arg for arg in args]
+            args = [self.quote_first_command_arg(sys.executable)] + sys.argv
             new_environ = os.environ.copy()
-            new_environ[self._reloader_environ_key] = 'true'
+            if reloader:
+                new_environ[self._reloader_environ_key] = 'true'
+            else:
+                new_environ[self._monitor_environ_key] = 'true'
+            proc = None
             try:
-                exit_code = os.spawnve(os.P_WAIT, sys.executable,
-                                       args, new_environ)
-            except KeyboardInterrupt:
-                if self.verbose > 1:
-                    raise
-                return 1
-            if exit_code != 3:
-                return exit_code
+                try:
+                    _turn_sigterm_into_systemexit()
+                    proc = subprocess.Popen(args, env=new_environ)
+                    exit_code = proc.wait()
+                    proc = None
+                except KeyboardInterrupt:
+                    print '^C caught in monitor process'
+                    if self.verbose > 1:
+                        raise
+                    return 1
+            finally:
+                if (proc is not None
+                    and hasattr(os, 'kill')):
+                    import signal
+                    try:
+                        os.kill(proc.pid, signal.SIGTERM)
+                    except (OSError, IOError):
+                        pass
+                
+            if reloader:
+                # Reloader always exits with code 3; but if we are
+                # a monitor, any exit code will restart
+                if exit_code != 3:
+                    return exit_code
             if self.verbose > 0:
                 print '-'*20, 'Restarting', '-'*20
 
@@ -374,17 +457,23 @@ class ServeCommand(Command):
             
 class LazyWriter(object):
 
-    def __init__(self, filename):
+    """
+    File-like object that opens a file lazily when it is first written
+    to.
+    """
+
+    def __init__(self, filename, mode='w'):
         self.filename = filename
         self.fileobj = None
         self.lock = threading.Lock()
+        self.mode = mode
         
     def open(self):
         if self.fileobj is None:
             self.lock.acquire()
             try:
                 if self.fileobj is None:
-                    self.fileobj = open(self.filename, 'a')
+                    self.fileobj = open(self.filename, self.mode)
             finally:
                 self.lock.release()
         return self.fileobj
@@ -500,3 +589,16 @@ def _cleanup_ports(bound_addresses, maxtries=30, sleeptime=2):
         else:
             raise SystemExit('Timeout waiting for port.')
         sock.close()
+
+def _turn_sigterm_into_systemexit():
+    """
+    Attempts to turn a SIGTERM exception into a SystemExit exception.
+    """
+    try:
+        import signal
+    except ImportError:
+        return
+    def handle_term(signo, frame):
+        raise SystemExit
+    signal.signal(signal.SIGTERM, handle_term)
+    
